@@ -31,7 +31,7 @@
 
 int Assemble(struct supertype *st, char *mddev, int mdfd,
 	     mddev_ident_t ident, char *conffile,
-	     mddev_dev_t devlist,
+	     mddev_dev_t devlist, char *backup_file,
 	     int readonly, int runstop,
 	     char *update,
 	     int verbose, int force)
@@ -117,7 +117,6 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	unsigned int num_devs;
 	mddev_dev_t tmpdev;
 	struct mdinfo info;
-	struct mddev_ident_s ident2;
 	char *avail;
 	int nextspare = 0;
 	
@@ -195,7 +194,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			super = NULL;
 		}
 		
-		dfd = open(devname, O_RDONLY|O_EXCL, 0);
+		dfd = dev_open(devname, O_RDONLY|O_EXCL);
 		if (dfd < 0) {
 			if ((inargv && verbose >= 0) || verbose > 0)
 				fprintf(stderr, Name ": cannot open device %s: %s\n",
@@ -204,11 +203,9 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			/* Impossible! */
 			fprintf(stderr, Name ": fstat failed for %s: %s\n",
 				devname, strerror(errno));
-			close(dfd);
 		} else if ((stb.st_mode & S_IFMT) != S_IFBLK) {
 			fprintf(stderr, Name ": %s is not a block device.\n",
 				devname);
-			close(dfd);
 		} else if (!tst && (tst = guess_super(dfd)) == NULL) {
 			if ((inargv && verbose >= 0) || verbose > 0)
 				fprintf(stderr, Name ": no recogniseable superblock\n");
@@ -216,13 +213,12 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			if ((inargv && verbose >= 0) || verbose > 0)
 				fprintf( stderr, Name ": no RAID superblock on %s\n",
 					 devname);
-			close(dfd);
 		} else {
-			tst->ss->getinfo_super(&info, &ident2, super);
-			close(dfd);
+			tst->ss->getinfo_super(&info, super);
 		}
+		if (dfd >= 0) close(dfd);
 
-		if (ident->uuid_set &&
+		if (ident->uuid_set && (!update || strcmp(update, "uuid")!= 0) &&
 		    (!super || same_uuid(info.uuid, ident->uuid, tst->ss->swapuuid)==0)) {
 			if ((inargv && verbose >= 0) || verbose > 0)
 				fprintf(stderr, Name ": %s has wrong uuid.\n",
@@ -230,7 +226,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			continue;
 		}
 		if (ident->name[0] &&
-		    (!super || strncmp(ident2.name, ident->name, 32)!=0)) {
+		    (!super || strncmp(info.name, ident->name, 32)!=0)) {
 			if ((inargv && verbose >= 0) || verbose > 0)
 				fprintf(stderr, Name ": %s has wrong name.\n",
 					devname);
@@ -269,6 +265,8 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			free(first_super);
 			return 1;
 		}
+
+
 		st = tst; /* commit to this format, if haven't already */
 		if (st->ss->compare_super(&first_super, super)) {
 			fprintf(stderr, Name ": superblock on %s doesn't match others - assembly aborted\n",
@@ -284,10 +282,24 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			struct stat stb2;
 			fstat(mdfd, &stb2);
 			info.array.md_minor = minor(stb2.st_rdev);
-			
+
+			if (strcmp(update, "uuid")==0 &&
+			    !ident->uuid_set) {
+				int rfd;
+				if ((rfd = open("/dev/urandom", O_RDONLY)) < 0 ||
+				    read(rfd, ident->uuid, 16) != 16) {
+					*(__u32*)(ident->uuid) = random();
+					*(__u32*)(ident->uuid+1) = random();
+					*(__u32*)(ident->uuid+2) = random();
+					*(__u32*)(ident->uuid+3) = random();
+				}
+				if (rfd >= 0) close(rfd);
+				ident->uuid_set = 1;
+			}
+			memcpy(info.uuid, ident->uuid, 16);
 			st->ss->update_super(&info, super, update, devname, verbose);
 			
-			dfd = open(devname, O_RDWR|O_EXCL, 0);
+			dfd = dev_open(devname, O_RDWR|O_EXCL);
 			if (dfd < 0) 
 				fprintf(stderr, Name ": Cannot open %s for superblock update\n",
 					devname);
@@ -361,7 +373,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		return 1;
 	}
 
-	st->ss->getinfo_super(&info, &ident2, first_super);
+	st->ss->getinfo_super(&info, first_super);
 
 	/* now we have some devices that might be suitable.
 	 * I wonder how many
@@ -372,7 +384,9 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	sparecnt=0;
 	for (i=0; i< bestcnt ;i++) {
 		int j = best[i];
-		int event_margin = !force;
+		int event_margin = 1; /* always allow a difference of '1'
+				       * like the kernel does
+				       */
 		if (j < 0) continue;
 		/* note: we ignore error flags in multipath arrays
 		 * as they don't make sense
@@ -418,7 +432,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 				devices[chosen_drive].devname, devices[chosen_drive].raid_disk,
 				(int)(devices[chosen_drive].events),
 				(int)(devices[most_recent].events));
-		fd = open(devices[chosen_drive].devname, O_RDWR|O_EXCL);
+		fd = dev_open(devices[chosen_drive].devname, O_RDWR|O_EXCL);
 		if (fd < 0) {
 			fprintf(stderr, Name ": Couldn't open %s for write - not updating\n",
 				devices[chosen_drive].devname);
@@ -468,7 +482,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		if (!devices[j].uptodate)
 			continue;
 		chosen_drive = j;
-		if ((fd=open(devices[j].devname, O_RDONLY|O_EXCL))< 0) {
+		if ((fd=dev_open(devices[j].devname, O_RDONLY|O_EXCL))< 0) {
 			fprintf(stderr, Name ": Cannot open %s: %s\n",
 				devices[j].devname, strerror(errno));
 			return 1;
@@ -485,7 +499,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		fprintf(stderr, Name ": No suitable drives found for %s\n", mddev);
 		return 1;
 	}
-	st->ss->getinfo_super(&info, &ident2, super);
+	st->ss->getinfo_super(&info, super);
 	for (i=0; i<bestcnt; i++) {
 		int j = best[i];
 		unsigned int desired_state;
@@ -534,9 +548,9 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 
 	if (change) {
 		int fd;
-		fd = open(devices[chosen_drive].devname, O_RDWR|O_EXCL);
+		fd = dev_open(devices[chosen_drive].devname, O_RDWR|O_EXCL);
 		if (fd < 0) {
-			fprintf(stderr, Name ": Could open %s for write - cannot Assemble array.\n",
+			fprintf(stderr, Name ": Could not open %s for write - cannot Assemble array.\n",
 				devices[chosen_drive].devname);
 			return 1;
 		}
@@ -549,6 +563,39 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		close(fd);
 	}
 
+	/* If we are in the middle of a reshape we may need to restore saved data
+	 * that was moved aside due to the reshape overwriting live data
+	 * The code of doing this lives in Grow.c
+	 */
+#ifndef MDASSEMBLE
+	if (info.reshape_active) {
+		int err = 0;
+		int *fdlist = malloc(sizeof(int)* bestcnt);
+		for (i=0; i<bestcnt; i++) {
+			int j = best[i];
+			if (j >= 0) {
+				fdlist[i] = dev_open(devices[j].devname, O_RDWR|O_EXCL);
+				if (fdlist[i] < 0) {
+					fprintf(stderr, Name ": Could not open %s for write - cannot Assemble array.\n",
+						devices[j].devname);
+					err = 1;
+					break;
+				}
+			} else
+				fdlist[i] = -1;
+		}
+		if (!err)
+			err = Grow_restart(st, &info, fdlist, bestcnt, backup_file);
+		while (i>0) {
+			i--;
+			if (fdlist[i]>=0) close(fdlist[i]);
+		}
+		if (err) {
+			fprintf(stderr, Name ": Failed to restore critical section for reshape, sorry.\n");
+			return err;
+		}
+	}
+#endif
 	/* count number of in-sync devices according to the superblock.
 	 * We must have this number to start the array without -s or -R
 	 */
@@ -576,8 +623,22 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 				fprintf(stderr, Name ": SET_BITMAP_FILE failed.\n");
 				return 1;
 			}
+		} else if (ident->bitmap_file) {
+			/* From config file */
+			int bmfd = open(ident->bitmap_file, O_RDWR);
+			if (bmfd < 0) {
+				fprintf(stderr, Name ": Could not open bitmap file %s\n",
+					ident->bitmap_file);
+				return 1;
+			}
+			if (ioctl(mdfd, SET_BITMAP_FILE, bmfd) != 0) {
+				fprintf(stderr, Name ": Failed to set bitmapfile for %s\n", mddev);
+				close(bmfd);
+				return 1;
+			}
+			close(bmfd);
 		}
-					
+
 		/* First, add the raid disks, but add the chosen one last */
 		for (i=0; i<= bestcnt; i++) {
 			int j;
