@@ -53,7 +53,8 @@ int Detail(char *dev, int brief, int test)
 	int is_rebuilding = 0;
 	int failed = 0;
 	struct supertype *st = NULL;
-	int max_disks = MD_SB_DISKS;
+	int max_disks = MD_SB_DISKS; /* just a default */
+	struct mdinfo info;
 
 	void *super = NULL;
 	int rv = test ? 4 : 1;
@@ -105,17 +106,15 @@ int Detail(char *dev, int brief, int test)
 		    disk.major == 0 &&
 		    disk.minor == 0)
 			continue;
-		if ((dv=map_dev(disk.major, disk.minor))) {
+		if ((dv=map_dev(disk.major, disk.minor, 1))) {
 			if (!super && (disk.state & (1<<MD_DISK_ACTIVE))) {
 				/* try to read the superblock from this device
 				 * to get more info
 				 */
-				int fd2 = open(dv, O_RDONLY);
+				int fd2 = dev_open(dv, O_RDONLY);
 				if (fd2 >=0 && st &&
 				    st->ss->load_super(st, fd2, &super, NULL) == 0) {
-					struct mdinfo info;
-					struct mddev_ident_s ident;
-					st->ss->getinfo_super(&info, &ident, super);
+					st->ss->getinfo_super(&info, super);
 					if (info.array.ctime != array.ctime ||
 					    info.array.level != array.level) {
 						free(super);
@@ -132,9 +131,10 @@ int Detail(char *dev, int brief, int test)
 	if (brief) 
 		printf("ARRAY %s level=%s num-devices=%d", dev, c?c:"-unknown-",array.raid_disks );
 	else {
+		mdu_bitmap_file_t bmf;
 		unsigned long array_size;
 		unsigned long long larray_size;
-		struct mdstat_ent *ms = mdstat_read(0);
+		struct mdstat_ent *ms = mdstat_read(0, 0);
 		struct mdstat_ent *e;
 		int devnum = array.md_minor;
 		if (major(stb.st_rdev) != MD_MAJOR)
@@ -163,14 +163,31 @@ int Detail(char *dev, int brief, int test)
 		printf("     Raid Level : %s\n", c?c:"-unknown-");
 		if (larray_size)
 			printf("     Array Size : %llu%s\n", (larray_size>>10), human_size(larray_size));
-		if (array.level >= 1)
-			printf("    Device Size : %d%s\n", array.size, human_size((long long)array.size<<10));
+		if (array.level >= 1) {
+			if (array.major_version != 0 &&
+			    (larray_size >= 0xFFFFFFFFULL|| array.size == 0)) {
+				unsigned long long dsize = get_component_size(fd);
+				if (dsize > 0)
+					printf("    Device Size : %llu%s\n", dsize, human_size((long long)array.size<<10));
+				else
+					printf("    Device Size : unknown\n");
+			} else
+				printf("    Device Size : %d%s\n", array.size, human_size((long long)array.size<<10));
+		}
 		printf("   Raid Devices : %d\n", array.raid_disks);
 		printf("  Total Devices : %d\n", array.nr_disks);
 		printf("Preferred Minor : %d\n", array.md_minor);
 		printf("    Persistence : Superblock is %spersistent\n",
 		       array.not_persistent?"not ":"");
 		printf("\n");
+		/* Only try GET_BITMAP_FILE for 0.90.01 and later */
+		if (vers >= 9001 &&
+		    ioctl(fd, GET_BITMAP_FILE, &bmf) == 0 &&
+		    bmf.pathname[0]) {
+			printf("  Intent Bitmap : %s\n", bmf.pathname);
+			printf("\n");
+		} else if (array.state & (1<<MD_SB_BITMAP_PRESENT))
+			printf("  Intent Bitmap : Internal\n\n");
 		atime = array.utime;
 		printf("    Update Time : %.24s\n", ctime(&atime));
 		printf("          State : %s%s%s\n",
@@ -206,11 +223,47 @@ int Detail(char *dev, int brief, int test)
 		}
 	
 		if (e && e->percent >= 0) {
-			printf(" Rebuild Status : %d%% complete\n\n", e->percent);
+			printf(" Re%s Status : %d%% complete\n",
+			       (super && info.reshape_active)? "shape":"build",
+			       e->percent);
 			is_rebuilding = 1;
 		}
 		free_mdstat(ms);
 
+		if (super && info.reshape_active) {
+#if 0
+This is pretty boring
+			printf("  Reshape pos'n : %llu%s\n", (unsigned long long) info.reshape_progress<<9,
+			       human_size(info.reshape_progress<<9));
+#endif
+			if (info.delta_disks > 0)
+				printf("  Delta Devices : %d, (%d->%d)\n",
+				       info.delta_disks, array.raid_disks - info.delta_disks, array.raid_disks);
+			if (info.delta_disks < 0)
+				printf("  Delta Devices : %d, (%d->%d)\n",
+				       info.delta_disks, array.raid_disks, array.raid_disks + info.delta_disks);
+			if (info.new_level != array.level) {
+				char *c = map_num(pers, info.new_level);
+				printf("      New Level : %s\n", c?c:"-unknown-");
+			}
+			if (info.new_level != array.level ||
+			    info.new_layout != array.layout) {
+				if (info.new_level == 5) {
+					char *c = map_num(r5layout, info.new_layout);
+					printf("     New Layout : %s\n",
+					       c?c:"-unknown-");
+				}
+				if (info.new_level == 10) {
+					printf("     New Layout : near=%d, far=%d\n",
+					       info.new_layout&255,
+					       (info.new_layout>>8)&255);
+				}
+			}
+			if (info.new_chunk != array.chunk_size)
+				printf("  New Chunksize : %dK\n", info.new_chunk/1024);
+			printf("\n");
+		} else if (e && e->percent >= 0)
+			printf("\n");
 		if (super && st)
 			st->ss->detail_super(super);
 
@@ -220,6 +273,7 @@ int Detail(char *dev, int brief, int test)
 	for (d=0; d<max_disks; d++) {
 		disks[d].state = (1<<MD_DISK_REMOVED);
 		disks[d].major = disks[d].minor = 0;
+		disks[d].number = disks[d].raid_disk = d;
 	}
 
 	next = array.raid_disks;
@@ -291,7 +345,7 @@ int Detail(char *dev, int brief, int test)
 				rv |= 2;
 			rv |= 1;
 		}
-		if ((dv=map_dev(disk.major, disk.minor))) {
+		if ((dv=map_dev(disk.major, disk.minor, 0))) {
 			if (brief) {
 				if (devices) {
 					devices = realloc(devices,
