@@ -40,17 +40,16 @@ int main(int argc, char *argv[])
 	int mode = 0;
 	int opt;
 	int option_index;
-	char *help_text;
 	char *c;
 	int rv;
 	int i;
 
 	int chunk = 0;
-	int size = -1;
+	long long size = -1;
 	int level = UnSet;
 	int layout = UnSet;
 	int raiddisks = 0;
-	int max_disks = MD_SB_DISKS;
+	int max_disks = MD_SB_DISKS; /* just a default */
 	int sparedisks = 0;
 	struct mddev_ident_s ident;
 	char *configfile = NULL;
@@ -63,6 +62,7 @@ int main(int argc, char *argv[])
 	int write_behind = 0;
 	int bitmap_fd = -1;
 	char *bitmap_file = NULL;
+	char *backup_file = NULL;
 	int bitmap_chunk = UnSet;
 	int SparcAdjust = 0;
 	mddev_dev_t devlist = NULL;
@@ -97,6 +97,7 @@ int main(int argc, char *argv[])
 	int dosyslog = 0;
 
 	int copies;
+	int print_help = 0;
 
 	int mdfd = -1;
 
@@ -111,6 +112,7 @@ int main(int argc, char *argv[])
 	ident.autof = 0;
 	ident.st = NULL;
 	ident.bitmap_fd = -1;
+	ident.bitmap_file = NULL;
 	ident.name[0] = 0;
 
 	while ((option_index = -1) ,
@@ -121,22 +123,12 @@ int main(int argc, char *argv[])
 		/* firstly, some mode-independant options */
 		switch(opt) {
 		case 'h':
-			help_text = Help;
 			if (option_index > 0 && 
 			    strcmp(long_options[option_index].name, "help-options")==0)
-				help_text = OptionHelp;
+				print_help = 2;
 			else
-				switch (mode) {
-				case ASSEMBLE : help_text = Help_assemble; break;
-				case BUILD    : help_text = Help_build; break;
-				case CREATE   : help_text = Help_create; break;
-				case MANAGE   : help_text = Help_manage; break;
-				case MISC     : help_text = Help_misc; break;
-				case MONITOR  : help_text = Help_monitor; break;
-				case GROW     : help_text = Help_grow; break;
-				}
-			fputs(help_text,stderr);
-			exit(0);
+				print_help = 1;
+			continue;
 
 		case 'V':
 			fputs(Version, stderr);
@@ -347,7 +339,7 @@ int main(int argc, char *argv[])
 			if (strcmp(optarg, "max")==0)
 				size = 0;
 			else {
-				size = strtol(optarg, &c, 10);
+				size = strtoll(optarg, &c, 10);
 				if (!optarg[0] || *c || size < 4) {
 					fprintf(stderr, Name ": invalid size: %s\n",
 						optarg);
@@ -779,11 +771,30 @@ int main(int argc, char *argv[])
 			ident.bitmap_fd = bitmap_fd; /* for Assemble */
 			continue;
 
+		case O(ASSEMBLE, 7):
+		case O(GROW, 7):
+			/* Specify a file into which grow might place a backup,
+			 * or from which assemble might recover a backup
+			 */
+			if (backup_file) {
+				fprintf(stderr, Name ": backup file already specified, rejecting %s\n", optarg);
+				exit(2);
+			}
+			backup_file = optarg;
+			continue;
+
 		case O(GROW,'b'):
 		case O(BUILD,'b'):
 		case O(CREATE,'b'): /* here we create the bitmap */
-			bitmap_file = optarg;
-			continue;
+			if (strcmp(optarg, "internal")== 0 ||
+			    strcmp(optarg, "none")== 0 ||
+			    strchr(optarg, '/') != NULL) {
+				bitmap_file = optarg;
+				continue;
+			}
+			/* probable typo */
+			fprintf(stderr, Name ": bitmap file must contain a '/', or be 'internal', or 'none'\n");
+			exit(2);
 
 		case O(GROW,4):
 		case O(BUILD,4):
@@ -815,12 +826,35 @@ int main(int argc, char *argv[])
 		/* We have now processed all the valid options. Anything else is
 		 * an error
 		 */
-		fprintf(stderr, Name ": option %c not valid in %s mode\n",
-			opt, map_num(modes, mode));
+		if (option_index > 0)
+			fprintf(stderr, Name ":option --%s not valid in %s mode\n",
+				long_options[option_index].name,
+				map_num(modes, mode));
+		else
+			fprintf(stderr, Name ": option -%c not valid in %s mode\n",
+				opt, map_num(modes, mode));
 		exit(2);
 
 	}
 
+	if (print_help) {
+		char *help_text = Help;
+		if (print_help == 2)
+			help_text = OptionHelp;
+		else
+			switch (mode) {
+			case ASSEMBLE : help_text = Help_assemble; break;
+			case BUILD    : help_text = Help_build; break;
+			case CREATE   : help_text = Help_create; break;
+			case MANAGE   : help_text = Help_manage; break;
+			case MISC     : help_text = Help_misc; break;
+			case MONITOR  : help_text = Help_monitor; break;
+			case GROW     : help_text = Help_grow; break;
+			}
+		fputs(help_text,stderr);
+		exit(0);
+	}
+
 	if (!mode && devs_found) {
 		mode = MISC;
 		devmode = 'Q';
@@ -912,20 +946,24 @@ int main(int argc, char *argv[])
 					rv |= 1;
 				else {
 					rv |= Assemble(ss, devlist->devname, mdfd, array_ident, configfile,
-						       NULL,
+						       NULL, backup_file,
 						       readonly, runstop, update, verbose-quiet, force);
 					close(mdfd);
 				}
 			}
 		} else if (!scan)
 			rv = Assemble(ss, devlist->devname, mdfd, &ident, configfile,
-				      devlist->next,
+				      devlist->next, backup_file,
 				      readonly, runstop, update, verbose-quiet, force);
 		else if (devs_found>0) {
 			if (update && devs_found > 1) {
 				fprintf(stderr, Name ": can only update a single array at a time\n");
 				exit(1);
 			}
+			if (backup_file && devs_found > 1) {
+				fprintf(stderr, Name ": can only assemble a single array when providing a backup file.\n");
+				exit(1);
+			}
 			for (dv = devlist ; dv ; dv=dv->next) {
 				mddev_ident_t array_ident = conf_get_ident(configfile, dv->devname);
 				if (array_ident == NULL) {
@@ -941,7 +979,7 @@ int main(int argc, char *argv[])
 					continue;
 				}
 				rv |= Assemble(ss, dv->devname, mdfd, array_ident, configfile,
-					       NULL,
+					       NULL, backup_file,
 					       readonly, runstop, update, verbose-quiet, force);
 				close(mdfd);
 			}
@@ -951,6 +989,14 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": No arrays found in config file\n");
 				rv = 1;
 			} else
+				if (update) {
+					fprintf(stderr, Name ": --update not meaningful with a --scan assembly.\n");
+					exit(1);
+				}
+				if (backup_file) {
+					fprintf(stderr, Name ": --backup_file not meaningful with a --scan assembly.\n");
+					exit(1);
+				}
 				for (; array_list; array_list = array_list->next) {
 					mdu_array_info_t array;
 					mdfd = open_mddev(array_list->devname, 
@@ -965,7 +1011,7 @@ int main(int argc, char *argv[])
 					else
 						rv |= Assemble(ss, array_list->devname, mdfd,
 							       array_list, configfile,
-							       NULL,
+							       NULL, NULL,
 							       readonly, runstop, NULL, verbose-quiet, force);
 					close(mdfd);
 				}
@@ -998,14 +1044,6 @@ int main(int argc, char *argv[])
 			rv = 1;
 			break;
 		}
-		if (ss == NULL) {
-			for(i=0; !ss && superlist[i]; i++) 
-				ss = superlist[i]->match_metadata_desc("default");
-		}
-		if (!ss) {
-			fprintf(stderr, Name ": internal error - no default metadata style\n");
-			exit(2);
-		}
 
 		rv = Create(ss, devlist->devname, mdfd, chunk, level, layout, size<0 ? 0 : size,
 			    raiddisks, sparedisks, ident.name,
@@ -1030,7 +1068,7 @@ int main(int argc, char *argv[])
 			if (devlist == NULL) {
 				if (devmode=='D' && scan) {
 					/* apply --detail to all devices in /proc/mdstat */
-					struct mdstat_ent *ms = mdstat_read(0);
+					struct mdstat_ent *ms = mdstat_read(0, 1);
 					struct mdstat_ent *e;
 					for (e=ms ; e ; e=e->next) {
 						char *name = get_md_name(e->devnum);
@@ -1051,7 +1089,7 @@ int main(int argc, char *argv[])
 					int progress=1, err;
 					int last = 0;
 					do {
-						struct mdstat_ent *ms = mdstat_read(0);
+						struct mdstat_ent *ms = mdstat_read(0, 0);
 						struct mdstat_ent *e;
 
 						if (!progress) last = 1;
@@ -1087,7 +1125,7 @@ int main(int argc, char *argv[])
 				case 'D':
 					rv |= Detail(dv->devname, brief?1+verbose:0, test); continue;
 				case 'K': /* Zero superblock */
-					rv |= Kill(dv->devname, force); continue;
+					rv |= Kill(dv->devname, force, quiet); continue;
 				case 'Q':
 					rv |= Query(dv->devname); continue;
 				case 'X':
@@ -1148,11 +1186,12 @@ int main(int argc, char *argv[])
 		} else if (layout != UnSet)
 			rv = Manage_reconfig(devlist->devname, mdfd, layout);
 		else if (size >= 0 || raiddisks)
-			rv = Manage_resize(devlist->devname, mdfd, size, raiddisks);
+			rv = Grow_reshape(devlist->devname, mdfd, quiet, backup_file,
+					  size, level, layout, chunk, raiddisks);
 		else if (bitmap_file) {
 			if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
 			rv = Grow_addbitmap(devlist->devname, mdfd, bitmap_file,
-					    bitmap_chunk, delay, write_behind);
+					    bitmap_chunk, delay, write_behind, force);
 		} else
 			fprintf(stderr, Name ": no changes to --grow\n");
 		break;
