@@ -28,7 +28,6 @@
  */
 
 #include "mdadm.h"
-#include <asm/byteorder.h>
 
 /*
  * All handling for the 0.90.0 version superblock is in
@@ -81,7 +80,7 @@ void super0_swap_endian(struct mdp_superblock_s *sb)
 	sb->cp_events_hi = sb->cp_events_lo;
 	sb->cp_events_lo = t32;
 
-}	
+}
 
 #ifndef MDASSEMBLE
 
@@ -105,12 +104,51 @@ static void examine_super0(void *sbv)
 	printf("  Creation Time : %.24s\n", ctime(&atime));
 	c=map_num(pers, sb->level);
 	printf("     Raid Level : %s\n", c?c:"-unknown-");
-	if ((int)sb->level >= 0)
+	if ((int)sb->level >= 0) {
+		int ddsks=0;
 		printf("    Device Size : %d%s\n", sb->size, human_size((long long)sb->size<<10));
+		switch(sb->level) {
+		case 1: ddsks=1;break;
+		case 4:
+		case 5: ddsks = sb->raid_disks-1; break;
+		case 6: ddsks = sb->raid_disks-2; break;
+		case 10: ddsks = sb->raid_disks / (sb->layout&255) / ((sb->layout>>8)&255);
+		}
+		if (ddsks)
+			printf("     Array Size : %llu%s\n", (unsigned long long)ddsks * sb->size,
+			       human_size(ddsks*(long long)sb->size<<10));
+	}
 	printf("   Raid Devices : %d\n", sb->raid_disks);
 	printf("  Total Devices : %d\n", sb->nr_disks);
 	printf("Preferred Minor : %d\n", sb->md_minor);
 	printf("\n");
+	if (sb->minor_version > 90 && (sb->reshape_position+1) != 0) {
+		printf("  Reshape pos'n : %llu%s\n", (unsigned long long)sb->reshape_position/2, human_size((long long)sb->reshape_position<<9));
+		if (sb->delta_disks) {
+			printf("  Delta Devices : %d", sb->delta_disks);
+			if (sb->delta_disks)
+				printf(" (%d->%d)\n", sb->raid_disks-sb->delta_disks, sb->raid_disks);
+			else
+				printf(" (%d->%d)\n", sb->raid_disks, sb->raid_disks+sb->delta_disks);
+		}
+		if (sb->new_level != sb->level) {
+			c = map_num(pers, sb->new_level);
+			printf("      New Level : %s\n", c?c:"-unknown-");
+		}
+		if (sb->new_layout != sb->layout) {
+			if (sb->level == 5) {
+				c = map_num(r5layout, sb->new_layout);
+				printf("     New Layout : %s\n", c?c:"-unknown-");
+			}
+			if (sb->level == 10) {
+				printf("     New Layout : near=%d, far=%d\n",
+				       sb->new_layout&255, (sb->new_layout>>8)&255);
+			}
+		}
+		if (sb->new_chunk != sb->chunk_size)
+			printf("  New Chunksize : %d\n", sb->new_chunk);
+		printf("\n");
+	}
 	atime = sb->utime;
 	printf("    Update Time : %.24s\n", ctime(&atime));
 	printf("          State : %s\n",
@@ -144,7 +182,7 @@ static void examine_super0(void *sbv)
 	case -1:
 		printf("       Rounding : %dK\n", sb->chunk_size/1024);
 		break;
-	default: break;		
+	default: break;
 	}
 	printf("\n");
 	printf("      Number   Major   Minor   RaidDevice State\n");
@@ -166,7 +204,7 @@ static void examine_super0(void *sbv)
 		if (dp->state & (1<<MD_DISK_REMOVED)) printf(" removed");
 		if (wonly) printf(" write-mostly");
 		if (dp->state == 0) printf(" spare");
-		if ((dv=map_dev(dp->major, dp->minor)))
+		if ((dv=map_dev(dp->major, dp->minor, 0)))
 			printf("   %s", dv);
 		printf("\n");
 		if (d == -1) printf("\n");
@@ -227,7 +265,7 @@ static void uuid_from_super0(int uuid[4], void * sbv)
 	}
 }
 
-static void getinfo_super0(struct mdinfo *info, mddev_ident_t ident, void *sbv)
+static void getinfo_super0(struct mdinfo *info, void *sbv)
 {
 	mdp_super_t *sb = sbv;
 	int working = 0;
@@ -241,6 +279,9 @@ static void getinfo_super0(struct mdinfo *info, mddev_ident_t ident, void *sbv)
 	info->array.layout = sb->layout;
 	info->array.md_minor = sb->md_minor;
 	info->array.ctime = sb->ctime;
+	info->array.utime = sb->utime;
+	info->array.chunk_size = sb->chunk_size;
+	info->component_size = sb->size*2;
 
 	info->disk.state = sb->this_disk.state;
 	info->disk.major = sb->this_disk.major;
@@ -249,10 +290,21 @@ static void getinfo_super0(struct mdinfo *info, mddev_ident_t ident, void *sbv)
 	info->disk.number = sb->this_disk.number;
 
 	info->events = md_event(sb);
+	info->data_offset = 0;
 
 	uuid_from_super0(info->uuid, sbv);
 
-	ident->name[0] = 0;
+	if (sb->minor_version > 90 && (sb->reshape_position+1) != 0) {
+		info->reshape_active = 1;
+		info->reshape_progress = sb->reshape_position;
+		info->new_level = sb->new_level;
+		info->delta_disks = sb->delta_disks;
+		info->new_layout = sb->new_layout;
+		info->new_chunk = sb->new_chunk;
+	} else
+		info->reshape_active = 0;
+
+	info->name[0] = 0;
 	/* work_disks is calculated rather than read directly */
 	for (i=0; i < MD_SB_DISKS; i++)
 		if ((sb->disks[i].state & (1<<MD_DISK_SYNC)) &&
@@ -365,6 +417,8 @@ static int update_super0(struct mdinfo *info, void *sbv, char *update, char *dev
 		sb->set_uuid2 = info->uuid[2];
 		sb->set_uuid3 = info->uuid[3];
 	}
+	if (strcmp(update, "_reshape_progress")==0)
+		sb->reshape_position = info->reshape_progress;
 
 	sb->sb_csum = calc_sb0_csum(sb);
 	return rv;
@@ -378,7 +432,7 @@ static __u64 event_super0(void *sbv)
 
 
 
-static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info, char *ignored_name)
+static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info, unsigned long long size, char *ignored_name)
 {
 	mdp_super_t *sb = malloc(MD_SB_BYTES + sizeof(bitmap_super_t));
 	int spares;
@@ -407,6 +461,8 @@ static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info,
 		sb->set_uuid0 = random();
 	sb->ctime = time(0);
 	sb->level = info->level;
+	if (size != info->size)
+		return 0;
 	sb->size = info->size;
 	sb->nr_disks = info->nr_disks;
 	sb->raid_disks = info->raid_disks;
@@ -441,7 +497,7 @@ static void add_to_super0(void *sbv, mdu_disk_info_t *dinfo)
 {
 	mdp_super_t *sb = sbv;
 	mdp_disk_t *dk = &sb->disks[dinfo->number];
-	
+
 	dk->number = dinfo->number;
 	dk->major = dinfo->major;
 	dk->minor = dinfo->minor;
@@ -466,9 +522,9 @@ static int store_super0(struct supertype *st, int fd, void *sbv)
 			dsize = ((unsigned long long)size)<<9;
 	}
 
-	if (dsize < MD_RESERVED_SECTORS*2)
+	if (dsize < MD_RESERVED_SECTORS*2*512)
 		return 2;
-	
+
 	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
 
 	offset *= 512;
@@ -582,7 +638,7 @@ static int load_super0(struct supertype *st, int fd, void **sbp, char *devname)
 				devname, size);
 		return 1;
 	}
-	
+
 	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
 
 	offset *= 512;
@@ -677,7 +733,7 @@ static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int
 	mdp_super_t *sb = sbv;
 	bitmap_super_t *bms = (bitmap_super_t*)(((char*)sb) + MD_SB_BYTES);
 
-	
+
 	min_chunk = 4096; /* sub-page chunks don't work yet.. */
 	bits = (size * 512)/ min_chunk +1;
 	while (bits > max_bits) {
@@ -691,7 +747,7 @@ static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int
 
 	sb->state |= (1<<MD_SB_BITMAP_PRESENT);
 
-	memset(bms, sizeof(*bms), 0);
+	memset(bms, 0, sizeof(*bms));
 	bms->magic = __cpu_to_le32(BITMAP_MAGIC);
 	bms->version = __cpu_to_le32(major);
 	uuid_from_super0((int*)bms->uuid, sb);
@@ -704,7 +760,7 @@ static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int
 
 	return 1;
 }
-		
+
 
 void locate_bitmap0(struct supertype *st, int fd, void *sbv)
 {
@@ -723,7 +779,7 @@ void locate_bitmap0(struct supertype *st, int fd, void *sbv)
 
 	if (dsize < MD_RESERVED_SECTORS*2)
 		return;
-	
+
 	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
 
 	offset *= 512;
@@ -756,8 +812,8 @@ int write_bitmap0(struct supertype *st, int fd, void *sbv)
 	}
 
 	if (dsize < MD_RESERVED_SECTORS*2)
-		return -1;
-	
+	return -1;
+
 	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
 
 	offset *= 512;
@@ -766,7 +822,9 @@ int write_bitmap0(struct supertype *st, int fd, void *sbv)
 		return 3;
 
 
-	write(fd, ((char*)sb)+MD_SB_BYTES, sizeof(bitmap_super_t));
+	if (write(fd, ((char*)sb)+MD_SB_BYTES, sizeof(bitmap_super_t)) !=
+	    sizeof(bitmap_super_t))
+		return -2;
 	towrite = 64*1024 - MD_SB_BYTES - sizeof(bitmap_super_t);
 	memset(buf, 0xff, sizeof(buf));
 	while (towrite > 0) {
