@@ -28,8 +28,6 @@
  */
 
 #include "mdadm.h"
-#include <endian.h>
-#include "asm/byteorder.h"
 /*
  * The version-1 superblock :
  * All numeric fields are little-endian.
@@ -58,7 +56,14 @@ struct mdp_superblock_1 {
 				 * NOTE: signed, so bitmap can be before superblock
 				 * only meaningful of feature_map[0] is set.
 				 */
-	__u8	pad1[128-100];	/* set to 0 when written */
+
+	/* These are only valid with feature bit '4' */
+	__u32	new_level;	/* new level we are reshaping to		*/
+	__u64	reshape_position;	/* next address in array-space for reshape */
+	__u32	delta_disks;	/* change in number of raid_disks		*/
+	__u32	new_layout;	/* new layout					*/
+	__u32	new_chunk;	/* new chunk size (bytes)			*/
+	__u8	pad1[128-124];	/* set to 0 when written */
 
 	/* constant this-device information - 64 bytes */
 	__u64	data_offset;	/* sector start of data, often 0 */
@@ -89,6 +94,15 @@ struct mdp_superblock_1 {
 	__u16	dev_roles[0];	/* role in array, or 0xffff for a spare, or 0xfffe for faulty */
 };
 
+/* feature_map bits */
+#define MD_FEATURE_BITMAP_OFFSET	1
+#define	MD_FEATURE_RECOVERY_OFFSET	2 /* recovery_offset is present and
+					   * must be honoured
+					   */
+#define	MD_FEATURE_RESHAPE_ACTIVE	4
+
+#define	MD_FEATURE_ALL			(1|2|4)
+
 #ifndef offsetof
 #define offsetof(t,f) ((int)&(((t*)0)->f))
 #endif
@@ -132,7 +146,8 @@ static void examine_super1(void *sbv)
 	char *c;
 
 	printf("          Magic : %08x\n", __le32_to_cpu(sb->magic));
-	printf("        Version : %02d.%02d\n", 1, __le32_to_cpu(sb->feature_map));
+	printf("        Version : %02d\n", 1);
+	printf("    Feature Map : 0x%x\n", __le32_to_cpu(sb->feature_map));
 	printf("     Array UUID : ");
 	for (i=0; i<16; i++) {
 		if ((i&3)==0 && i != 0) printf(":");
@@ -148,10 +163,26 @@ static void examine_super1(void *sbv)
 	printf("   Raid Devices : %d\n", __le32_to_cpu(sb->raid_disks));
 	printf("\n");
 	printf("    Device Size : %llu%s\n", (unsigned long long)sb->data_size, human_size(sb->data_size<<9));
+	if (__le32_to_cpu(sb->level) >= 0) {
+		int ddsks=0;
+		switch(__le32_to_cpu(sb->level)) {
+		case 1: ddsks=1;break;
+		case 4:
+		case 5: ddsks = sb->raid_disks-1; break;
+		case 6: ddsks = sb->raid_disks-2; break;
+		case 10: ddsks = sb->raid_disks / (sb->layout&255) / ((sb->layout>>8)&255);
+		}
+		if (ddsks)
+			printf("     Array Size : %llu%s\n", ddsks*(unsigned long long)sb->size, human_size(ddsks*sb->size<<9));
+		if (sb->size != sb->data_size)
+			printf("      Used Size : %llu%s\n", (unsigned long long)sb->size, human_size(sb->size<<9));
+	}
 	if (sb->data_offset)
 		printf("    Data Offset : %llu sectors\n", (unsigned long long)__le64_to_cpu(sb->data_offset));
 	if (sb->super_offset)
 		printf("   Super Offset : %llu sectors\n", (unsigned long long)__le64_to_cpu(sb->super_offset));
+	if (__le32_to_cpu(sb->feature_map) & MD_FEATURE_RECOVERY_OFFSET)
+		printf("Recovery Offset : %llu sectors\n", (unsigned long long)__le64_to_cpu(sb->recovery_offset));
 	printf("          State : %s\n", (__le64_to_cpu(sb->resync_offset)+1)? "active":"clean");
 	printf("    Device UUID : ");
 	for (i=0; i<16; i++) {
@@ -159,6 +190,43 @@ static void examine_super1(void *sbv)
 		printf("%02x", sb->device_uuid[i]);
 	}
 	printf("\n");
+	printf("\n");
+	if (sb->feature_map & __cpu_to_le32(MD_FEATURE_BITMAP_OFFSET)) {
+		printf("Internal Bitmap : %ld sectors from superblock\n",
+		       (long)__le32_to_cpu(sb->bitmap_offset));
+	}
+	if (sb->feature_map & __le32_to_cpu(MD_FEATURE_RESHAPE_ACTIVE)) {
+		printf("  Reshape pos'n : %llu%s\n", (unsigned long long)__le64_to_cpu(sb->reshape_position)/2,
+		       human_size(__le64_to_cpu(sb->reshape_position)<<9));
+		if (__le32_to_cpu(sb->delta_disks)) {
+			printf("  Delta Devices : %d", __le32_to_cpu(sb->delta_disks));
+			if (__le32_to_cpu(sb->delta_disks))
+				printf(" (%d->%d)\n",
+				       __le32_to_cpu(sb->raid_disks)-__le32_to_cpu(sb->delta_disks),
+				       __le32_to_cpu(sb->raid_disks));
+			else
+				printf(" (%d->%d)\n", __le32_to_cpu(sb->raid_disks),
+				       __le32_to_cpu(sb->raid_disks)+__le32_to_cpu(sb->delta_disks));
+		}
+		if (__le32_to_cpu(sb->new_level) != __le32_to_cpu(sb->level)) {
+			c = map_num(pers, __le32_to_cpu(sb->new_level));
+			printf("      New Level : %s\n", c?c:"-unknown-");
+		}
+		if (__le32_to_cpu(sb->new_layout) != __le32_to_cpu(sb->layout)) {
+			if (__le32_to_cpu(sb->level) == 5) {
+				c = map_num(r5layout, __le32_to_cpu(sb->new_layout));
+				printf("     New Layout : %s\n", c?c:"-unknown-");
+			}
+			if (__le32_to_cpu(sb->level) == 10) {
+				printf("     New Layout : near=%d, far=%d\n",
+				       __le32_to_cpu(sb->new_layout)&255,
+				       (__le32_to_cpu(sb->new_layout)>>8)&255);
+			}
+		}
+		if (__le32_to_cpu(sb->new_chunk) != __le32_to_cpu(sb->chunksize))
+			printf("  New Chunksize : %dK\n", __le32_to_cpu(sb->new_chunk)/2);
+		printf("\n");
+	}
 	if (sb->devflags) {
 		printf("      Flags :");
 		if (sb->devflags & WriteMostly1)
@@ -189,10 +257,12 @@ static void examine_super1(void *sbv)
 	case 0:
 	case 4:
 	case 5:
-		printf("     Chunk Size : %dK\n", __le32_to_cpu(sb->chunksize/2));
+	case 6:
+	case 10:
+		printf("     Chunk Size : %dK\n", __le32_to_cpu(sb->chunksize)/2);
 		break;
 	case -1:
-		printf("       Rounding : %dK\n", __le32_to_cpu(sb->chunksize/2));
+		printf("       Rounding : %dK\n", __le32_to_cpu(sb->chunksize)/2);
 		break;
 	default: break;
 	}
@@ -283,7 +353,7 @@ static void uuid_from_super1(int uuid[4], void * sbv)
 		cuuid[i] = super->set_uuid[i];
 }
 
-static void getinfo_super1(struct mdinfo *info, mddev_ident_t ident, void *sbv)
+static void getinfo_super1(struct mdinfo *info, void *sbv)
 {
 	struct mdp_superblock_1 *sb = sbv;
 	int working = 0;
@@ -298,6 +368,11 @@ static void getinfo_super1(struct mdinfo *info, mddev_ident_t ident, void *sbv)
 	info->array.layout = __le32_to_cpu(sb->layout);
 	info->array.md_minor = -1;
 	info->array.ctime = __le64_to_cpu(sb->ctime);
+	info->array.utime = __le64_to_cpu(sb->utime);
+	info->array.chunk_size = __le32_to_cpu(sb->chunksize)*512;
+
+	info->data_offset = __le64_to_cpu(sb->data_offset);
+	info->component_size = __le64_to_cpu(sb->size);
 
 	info->disk.major = 0;
 	info->disk.minor = 0;
@@ -324,8 +399,18 @@ static void getinfo_super1(struct mdinfo *info, mddev_ident_t ident, void *sbv)
 
 	memcpy(info->uuid, sb->set_uuid, 16);
 
-	strncpy(ident->name, sb->set_name, 32);
-	ident->name[32] = 0;
+	strncpy(info->name, sb->set_name, 32);
+	info->name[32] = 0;
+
+	if (sb->feature_map & __le32_to_cpu(MD_FEATURE_RESHAPE_ACTIVE)) {
+		info->reshape_active = 1;
+		info->reshape_progress = __le64_to_cpu(sb->reshape_position);
+		info->new_level = __le32_to_cpu(sb->new_level);
+		info->delta_disks = __le32_to_cpu(sb->delta_disks);
+		info->new_layout = __le32_to_cpu(sb->new_layout);
+		info->new_chunk = __le32_to_cpu(sb->new_chunk)<<9;
+	} else
+		info->reshape_active = 0;
 
 	for (i=0; i< __le32_to_cpu(sb->max_dev); i++) {
 		role = __le16_to_cpu(sb->dev_roles[i]);
@@ -381,6 +466,10 @@ static int update_super1(struct mdinfo *info, void *sbv, char *update, char *dev
 		/* make sure resync happens */
 		sb->resync_offset = ~0ULL;
 	}
+	if (strcmp(update, "uuid") == 0)
+		memcpy(sb->set_uuid, info->uuid, 16);
+	if (strcmp(update, "_reshape_progress")==0)
+		sb->reshape_position = __cpu_to_le64(info->reshape_progress);
 
 	sb->sb_csum = calc_sb_1_csum(sb);
 	return rv;
@@ -393,7 +482,7 @@ static __u64 event_super1(void *sbv)
 	return __le64_to_cpu(sb->events);
 }
 
-static int init_super1(struct supertype *st, void **sbp, mdu_array_info_t *info, char *name)
+static int init_super1(struct supertype *st, void **sbp, mdu_array_info_t *info, unsigned long long size, char *name)
 {
 	struct mdp_superblock_1 *sb = malloc(1024 + sizeof(bitmap_super_t));
 	int spares;
@@ -432,7 +521,7 @@ static int init_super1(struct supertype *st, void **sbp, mdu_array_info_t *info,
 	sb->ctime = __cpu_to_le64((unsigned long long)time(0));
 	sb->level = __cpu_to_le32(info->level);
 	sb->layout = __cpu_to_le32(info->layout);
-	sb->size = __cpu_to_le64(info->size*2ULL);
+	sb->size = __cpu_to_le64(size*2ULL);
 	sb->chunksize = __cpu_to_le32(info->chunk_size>>9);
 	sb->raid_disks = __cpu_to_le32(info->raid_disks);
 
@@ -475,13 +564,21 @@ static int store_super1(struct supertype *st, int fd, void *sbv)
 	struct mdp_superblock_1 *sb = sbv;
 	unsigned long long sb_offset;
 	int sbsize;
-	long size;
+	unsigned long size;
+	unsigned long long dsize;
 
-	if (ioctl(fd, BLKGETSIZE, &size))
-		return 1;
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size))
+			return 1;
+		else
+			dsize = (unsigned long long)size;
+	} else
+		dsize >>= 9;
 
-
-	if (size < 24)
+	if (dsize < 24)
 		return 2;
 
 	/*
@@ -494,7 +591,7 @@ static int store_super1(struct supertype *st, int fd, void *sbv)
 	 */
 	switch(st->minor_version) {
 	case 0:
-		sb_offset = size;
+		sb_offset = dsize;
 		sb_offset -= 8*2;
 		sb_offset &= ~(4*2-1);
 		break;
@@ -540,7 +637,8 @@ static int write_init_super1(struct supertype *st, void *sbv,
 	int rfd;
 	int rv;
 
-	long size;
+	unsigned long size, space;
+	unsigned long long dsize, array_size;
 	long long sb_offset;
 
 
@@ -576,12 +674,18 @@ static int write_init_super1(struct supertype *st, void *sbv,
 		free(refsb);
 	}
 
-	if (ioctl(fd, BLKGETSIZE, &size)) {
-		close(fd);
-		return 1;
-	}
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size))
+			return 1;
+		else
+			dsize = size;
+	} else
+		dsize >>= 9;
 
-	if (size < 24) {
+	if (dsize < 24) {
 		close(fd);
 		return 2;
 	}
@@ -594,26 +698,40 @@ static int write_init_super1(struct supertype *st, void *sbv,
 	 * 0: At least 8K, but less than 12K, from end of device
 	 * 1: At start of device
 	 * 2: 4K from start of device.
+	 * Depending on the array size, we might leave extra space
+	 * for a bitmap.
 	 */
+	array_size = __le64_to_cpu(sb->size);
 	switch(st->minor_version) {
 	case 0:
-		sb_offset = size;
+		sb_offset = dsize;
 		sb_offset -= 8*2;
 		sb_offset &= ~(4*2-1);
 		sb->super_offset = __cpu_to_le64(sb_offset);
 		sb->data_offset = __cpu_to_le64(0);
-		sb->data_size = __cpu_to_le64(sb_offset);
+		if (sb_offset-64*2 >= array_size && array_size > 8*1024*1024*2)
+			sb->data_size = __cpu_to_le64(sb_offset-64*2);
+		else
+			sb->data_size = __cpu_to_le64(sb_offset);
 		break;
 	case 1:
 		sb->super_offset = __cpu_to_le64(0);
-		sb->data_offset = __cpu_to_le64(4*2); /* leave 4k for super and bitmap */
-		sb->data_size = __cpu_to_le64(size - 4*2);
+		if (dsize - 64*2 >= array_size && array_size > 8*1024*1024*2)
+			space = 64*2;
+		else
+			space = 4*2;
+		sb->data_offset = __cpu_to_le64(space); /* leave space for super and bitmap */
+		sb->data_size = __cpu_to_le64(dsize - space);
 		break;
 	case 2:
 		sb_offset = 4*2;
+		if (dsize - 4*2 - 64*2 >= array_size && array_size > 8*1024*1024*2)
+			space = 64*2;
+		else
+			space = 4*2;
 		sb->super_offset = __cpu_to_le64(sb_offset);
-		sb->data_offset = __cpu_to_le64(sb_offset+4*2);
-		sb->data_size = __cpu_to_le64(size - 4*2 - 4*2);
+		sb->data_offset = __cpu_to_le64(sb_offset+space);
+		sb->data_size = __cpu_to_le64(dsize - 4*2 - space);
 		break;
 	default:
 		return -EINVAL;
@@ -670,6 +788,7 @@ static int compare_super1(void **firstp, void *secondv)
 static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
 {
 	unsigned long size;
+	unsigned long long dsize;
 	unsigned long long sb_offset;
 	struct mdp_superblock_1 *super;
 
@@ -707,17 +826,24 @@ static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
 		st->ss = NULL;
 		return 2;
 	}
-	if (ioctl(fd, BLKGETSIZE, &size)) {
-		if (devname)
-			fprintf(stderr, Name ": cannot find device size for %s: %s\n",
-				devname, strerror(errno));
-		return 1;
-	}
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size)) {
+			if (devname)
+				fprintf(stderr, Name ": cannot find device size for %s: %s\n",
+					devname, strerror(errno));
+			return 1;
+		}
+		dsize = size;
+	} else
+		dsize >>= 9;
 
-	if (size < 24) {
+	if (dsize < 24) {
 		if (devname)
-			fprintf(stderr, Name ": %s is too small for md: size is %lu sectors.\n",
-				devname, size);
+			fprintf(stderr, Name ": %s is too small for md: size is %llu sectors.\n",
+				devname, dsize);
 		return 1;
 	}
 
@@ -731,7 +857,7 @@ static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
 	 */
 	switch(st->minor_version) {
 	case 0:
-		sb_offset = size;
+		sb_offset = dsize;
 		sb_offset -= 8*2;
 		sb_offset &= ~(4*2-1);
 		break;
@@ -800,7 +926,8 @@ static struct supertype *match_metadata_desc1(char *arg)
 	st->ss = &super1;
 	st->max_devs = 384;
 	if (strcmp(arg, "1") == 0 ||
-	    strcmp(arg, "1.0") == 0) {
+	    strcmp(arg, "1.0") == 0 ||
+	    strcmp(arg, "default/large") == 0) {
 		st->minor_version = 0;
 		return st;
 	}
@@ -826,6 +953,14 @@ static __u64 avail_size1(struct supertype *st, __u64 devsize)
 	if (devsize < 24)
 		return 0;
 
+	/* if the device is bigger than 8Gig, save 64k for bitmap usage,
+	 * if biffer than 200Gig, save 128k
+	 */
+	if (devsize > 200*1024*1024*2)
+		devsize -= 128*2;
+	else if (devsize > 8*1024*1024*2)
+		devsize -= 64*2;
+
 	switch(st->minor_version) {
 	case 0:
 		/* at end */
@@ -842,7 +977,8 @@ static __u64 avail_size1(struct supertype *st, __u64 devsize)
 
 static int
 add_internal_bitmap1(struct supertype *st, void *sbv,
-		     int chunk, int delay, int write_behind, int *sizep, int may_change)
+		     int chunk, int delay, int write_behind, unsigned long long size,
+		     int may_change, int major)
 {
 	/*
 	 * If not may_change, then this is a 'Grow', and the bitmap
@@ -851,10 +987,9 @@ add_internal_bitmap1(struct supertype *st, void *sbv,
 	 * before the superblock if we like, or may move the start.
 	 * For now, just squeeze the bitmap into 3k and don't change anything.
 	 *
-	 * size is in K,  chunk is in bytes !!!
+	 * size is in sectors,  chunk is in bytes !!!
 	 */
 
-	unsigned long long size = *sizep;
 	unsigned long long bits;
 	unsigned long long max_bits = (3*512 - sizeof(bitmap_super_t)) * 8;
 	unsigned long long min_chunk;
@@ -868,7 +1003,7 @@ add_internal_bitmap1(struct supertype *st, void *sbv,
 
 
 	min_chunk = 4096; /* sub-page chunks don't work yet.. */
-	bits = (size*1024)/min_chunk +1;
+	bits = (size*512)/min_chunk +1;
 	while (bits > max_bits) {
 		min_chunk *= 2;
 		bits = (bits+1)/2;
@@ -877,17 +1012,19 @@ add_internal_bitmap1(struct supertype *st, void *sbv,
 		chunk = min_chunk;
 	else if (chunk < min_chunk)
 		return 0; /* chunk size too small */
+	if (chunk == 0) /* rounding problem */
+		return 0;
 
 	sb->bitmap_offset = __cpu_to_le32(2);
 
 	sb->feature_map = __cpu_to_le32(__le32_to_cpu(sb->feature_map) | 1);
-	memset(bms, sizeof(*bms), 0);
+	memset(bms, 0, sizeof(*bms));
 	bms->magic = __cpu_to_le32(BITMAP_MAGIC);
-	bms->version = __cpu_to_le32(BITMAP_MAJOR);
+	bms->version = __cpu_to_le32(major);
 	uuid_from_super1((int*)bms->uuid, sb);
 	bms->chunksize = __cpu_to_le32(chunk);
 	bms->daemon_sleep = __cpu_to_le32(delay);
-	bms->sync_size = __cpu_to_le64(size<<1);
+	bms->sync_size = __cpu_to_le64(size);
 	bms->write_behind = __cpu_to_le32(write_behind);
 
 	return 1;
@@ -898,16 +1035,18 @@ void locate_bitmap1(struct supertype *st, int fd, void *sbv)
 {
 	unsigned long long offset;
 	struct mdp_superblock_1 *sb;
+	int mustfree = 0;
 
-	if (!sbv)
-		if (st->ss->load_super(st, fd, sbv, NULL))
+	if (!sbv) {
+		if (st->ss->load_super(st, fd, &sbv, NULL))
 			return; /* no error I hope... */
-
+		mustfree = 1;
+	}
 	sb = sbv;
 
 	offset = __le64_to_cpu(sb->super_offset);
 	offset += (long) __le32_to_cpu(sb->bitmap_offset);
-	if (!sbv)
+	if (mustfree)
 		free(sb);
 	lseek64(fd, offset<<9, 0);
 }
@@ -923,7 +1062,9 @@ int write_bitmap1(struct supertype *st, int fd, void *sbv)
 
 	locate_bitmap1(st, fd, sbv);
 
-	write(fd, ((char*)sb)+1024, sizeof(bitmap_super_t));
+	if (write(fd, ((char*)sb)+1024, sizeof(bitmap_super_t)) !=
+	    sizeof(bitmap_super_t))
+		return -2;
 	towrite = __le64_to_cpu(bms->sync_size) / (__le32_to_cpu(bms->chunksize)>>9);
 	towrite = (towrite+7) >> 3; /* bits to bytes */
 	memset(buf, 0xff, sizeof(buf));
